-----------------------
-- Scenario variable -- 
-----------------------

-- Bounding box of the arena, that only affects random spawning and clear_map
arena_box = {{-48,-48},{48,48}}

welcome_messages = {
	"On this map, everyone is in a different team, so you need to use /s <message> to be able to chat with other players",
	"Suggestions & feedback are welcome"
}

--	Default technologies applied to each player/force	
default_technologies = {	
	"bullet-speed-1",
	"bullet-speed-2",
	"bullet-speed-3",
	"bullet-speed-4",
	"bullet-speed-5",
	"bullet-speed-6",
	"shotgun-shell-speed-1",
	"shotgun-shell-speed-2",
	"shotgun-shell-speed-3",
	"bullet-damage-1",
	"bullet-damage-2",
	"bullet-damage-3",
	"bullet-damage-4",
	"bullet-damage-5",
	"shotgun-shell-damage-1",
	"shotgun-shell-damage-2"
}

-- Items found in chests, can handle has many tiers/groups as differents chests available (currently 7 max)
-- Probability must be uint
global.chest_loot_table = {
	["steel-chest"] = {
		{item = {name="land-mine",count=1},			probability = 3},
		{item = {name="grenade",count=2},			probability = 2},
		{item = {name="explosive-rocket",count=1},	probability = 2},
		{item = {name="firearm-magazine",count=5},	probability = 1},
		{item = {name="raw-fish",count=1},			probability = 1},
		{item = {name="shotgun-shell",count=2},		probability = 1}
	},
	["wooden-chest"] = {
		--{item = {name="piercing-rounds-magazine",count=5},	probability = 2}, -- too op
		{item = {name="cluster-grenade",count=1},			probability = 1},
		--{item = {name="piercing-shotgun-shell",count=2},	probability = 2}, -- too op
		{item = {name="poison-capsule",count=1},			probability = 1},
		{item = {name="gun-turret",count=1},				probability = 1},
		{item = {name="raw-fish",count=1},					probability = 2},
	}
}

-- Starting player items
player_default_items = {
	{name="submachine-gun", count=1},
	{name="firearm-magazine", count=30},
	{name="combat-shotgun",count=1},
	{name="shotgun-shell",count=10},
	{name="light-armor", count=1},
	{name="rocket-launcher",count=1},
	{name="grenade",count=2}
}

-- Add you spawn positions here. Useless if random_spawns=true 
-- Array of positions example : {{-26.5, -26.5}, {-0.5, -26.5}, {-0.5, 26.5}}
global.spawn_list = {} 

-- Activacte random spawning
random_spawns = true

-- Tile types that will not let players spawn on
random_spawn_exclude_tiles = {
	"concrete",
	"stone-path"
}

-- Number of spawns to be generated
global.random_spawn_limit = 100

-- Default turret life (default = 400) must be <= default
global.turret_default_health = 150

----------------------------------------
-- Changeable runtime via admin panel --
----------------------------------------

-- Kill limit for each round
global.kill_limit = 10

-- Delay (seconds) before a chest respawns
global.chest_respawn_delay = {
	["steel-chest"] = 3,
	["wooden-chest"] = 5
}

-- Initial delay (seconds) before filling chests
global.chest_initial_delay ={
	["steel-chest"] = 5,
	["wooden-chest"] = 10
}

-- Number of players displayed in the leaderboard gui
global.display_count = 5

------------------------------------
-- end of user-configurable datas --
------------------------------------

global.player_respawn_delay = 10
global.unique_player_count = 0
global.generated_spawn_list = {}
global.chest_respawn_ticks = {}
global.current_round_scores = {}
global.leaderboard_scores = {}
global.chest_table = {}
global.chest_initial_tick = {}
global.activated_chests = {}

script.on_init(function (event)	
	format_loot_table()
	game.surfaces[1].always_day = true
	chest_initial_tick_register(0)		
end)

----------------------
-- Round processing --
---------------------- 

-- triggeerd when any player's kill count reach kill_limit
function end_round()
	for _,p in pairs(game.connected_players) do
		-- clear player inventory
		p.clear_items_inside()
		-- show popup gui
		end_round_popup(p)
		-- lock player action
		if p.character then
			p.character.active=false
		end
	end
	-- add curscore to overall & clear curscore
	for i,v in pairs(global.current_round_scores) do
		global.leaderboard_scores[i] = global.leaderboard_scores[i] + v
		global.current_round_scores[i] = 0
	end
	-- clear chests content
	for _,chestlist in pairs(global.chest_table) do
		for i=1, #chestlist do
			chestlist[i].clear_items_inside()
		end
	end
	update_guis(true)
	-- reset tick table
	global.chest_initial_tick = {}
	global.chest_respawn_ticks = {}
	global.activated_chests = {}
	-- clear land mines/turrets
	clean_map()
end

-- triggered when a player respawns and 10s have passed since end_round() (when the last killed respawns)
function new_round()
	chest_initial_tick_register(game.tick)
	global.winner = nil
	global.wintick = nil
	-- unlock player, free popup gui, equip & tp
	for _,p in pairs(game.connected_players) do
		if p.gui.center.endframe then
			p.gui.center.endframe.destroy()
		end
		if p.character then
			p.character.active=true
		end
		prepare(p)
	end
end

--------------------
-- Event Handlers --
--------------------

script.on_event(defines.events.on_tick, function(event)	
	-- proceed chest refilling queue
	if global.chest_respawn_ticks[event.tick] then
		math.randomseed(event.tick)
		local chest = global.chest_respawn_ticks[event.tick]
		local loot = global.chest_loot_table[chest.name][math.random(#global.chest_loot_table[chest.name])]
		chest.insert(loot)
		global.chest_respawn_ticks[event.tick]=nil
	end
	-- proceed chest first filling
	if global.chest_initial_tick[event.tick] then
		local tier = global.chest_initial_tick[event.tick]
		global.activated_chests[tier] = true
		for i=1 ,#global.chest_table[tier] do 
			global.chest_table[tier][i].insert(global.chest_loot_table[tier][math.random(#global.chest_loot_table[tier])])
		end
		game.print("Chest activation : " .. tier)
		global.chest_respawn_ticks[event.tick]=nil
	end
end)

-- populate chest_table as chest are generated
script.on_event(defines.events.on_chunk_generated, function(event)
	local a = event.surface.find_entities_filtered{force="neutral", area=event.area}
    for _,c in pairs(a) do
		global.chest_table[c.name] = global.chest_table[c.name] or {}
        table.insert(global.chest_table[c.name],c)
    end
end)

script.on_event(defines.events.on_built_entity, function(event)
	if event.created_entity.name == "gun-turret" then
		event.created_entity.health = global.turret_default_health
	end
end)

script.on_event(defines.events.on_player_left_game, function (event)
  local player = game.players[event.player_index]
  -- save player score to overall & del curscore
  global.leaderboard_scores[player.name] = global.leaderboard_scores[player.name] + global.current_round_scores[player.name]
  global.current_round_scores[player.name] = nil
  -- remove player's land-mines
  clean_map(player)
  -- del force
  game.merge_forces(player.name, "player")
  --remove player from other's gui
  update_guis(true)
end)

-- whenever any player's inventory changes it means a chest had just been looted so we register a chest respawn
script.on_event(defines.events.on_player_main_inventory_changed, 	function (event)	chest_respawn_tick_register(event.tick) end)
script.on_event(defines.events.on_player_quickbar_inventory_changed,function (event)	chest_respawn_tick_register(event.tick) end)
script.on_event(defines.events.on_player_ammo_inventory_changed,	function (event)	chest_respawn_tick_register(event.tick) end)
script.on_event(defines.events.on_player_armor_inventory_changed,	function (event)	chest_respawn_tick_register(event.tick) end)
script.on_event(defines.events.on_player_gun_inventory_changed, 	function (event)	chest_respawn_tick_register(event.tick) end)

script.on_event(defines.events.on_player_joined_game, function (event)
	local player = game.players[event.player_index] 
	for _,m in pairs(welcome_messages) do
		player.print(m)
	end
	global.current_round_scores[player.name] = 0
	-- check new player or reconnected player
	if not global.leaderboard_scores[player.name] then
		global.leaderboard_scores[player.name] = 0
		global.unique_player_count = global.unique_player_count + 1
	end
	if player.force.name == "player" then
		--create force with same name as player
		local force = game.create_force(player.name)	
		--	apply technologies
		for _,t in pairs(default_technologies) do
			force.technologies[t].researched = true
		end
		-- switch player to this force
		player.force = force
	end
	--create player gui & update
	create_guis(player)	
	-- quick restart if a 2nd player just joined
	if (#game.connected_players == 2) then
		end_round()
		new_round()
	else
		prepare(player)
	end
end)

-- use to restart a round : when the last killed player from a previous round respawns then we can process the new round
script.on_event(defines.events.on_player_respawned, function (event)
	local player = game.players[event.player_index]
	-- check that the round has ended for the duration of the respawn time (10s)
	if global.wintick and event.tick > global.wintick + global.player_respawn_delay*60 - 50 then
		new_round()
	else
		prepare(player)
	end
end)

-- allow detection of killer but get also triggered by land mines/tower destruction
script.on_event(defines.events.on_entity_died, function (event)
	-- add score to killer
	if event.entity.force ~= event.force and event.entity.name == "player" then
		global.current_round_scores[event.force.name] = global.current_round_scores[event.force.name] + 1
	end
	-- check victory
	if global.current_round_scores[event.force.name] >= global.kill_limit then
		global.winner = event.force.name
		global.wintick = event.tick
		end_round() --will proceed full ui update
	else	
		update_guis()
	end
end)

script.on_event(defines.events.on_gui_click, function (event)
	if event.element.name=="showbutton" then
		event.element.parent.parent.adminpanel.style.visible = not event.element.parent.parent.adminpanel.style.visible
	elseif event.element.name=="updatebutton" then
		local g = event.element.parent.admintable.cfgtext
		if global.kill_limit ~= tonumber(g.kill_limit.text) then 
			global.kill_limit = tonumber(g.kill_limit.text)
			game.print("Kill limit per round is " ..g.kill_limit.text)
		end
		if global.display_count ~= tonumber(g.count.text) then
			global.display_count = tonumber(g.count.text)
			update_guis(true)
		end
		for tier,_ in pairs(global.chest_loot_table) do
			if global.chest_initial_delay[tier] ~= (tonumber(g[tier .. "_initdelay"].text) or global.chest_initial_delay[tier]) then
				global.chest_initial_delay[tier] = tonumber(g[tier .. "_initdelay"].text)
				game.print(tier .. " activation after " .. global.chest_respawn_delay[tier] .. " seconds")
			end
			if global.chest_respawn_delay[tier] ~= (tonumber(g[tier .. "_respawndelay"].text) or global.chest_respawn_delay[tier]) then
				global.chest_respawn_delay[tier] = tonumber(g[tier .. "_respawndelay"].text)
				game.print(tier .. " respawns after " .. global.chest_initial_delay[tier] .." seconds")
			end	
		end			
	elseif event.element.name=="restartbutton" then
		end_round()
		new_round()
	end
end)

-----------------
-- Gui related --
-----------------

-- create guis for the specified player (joining player)
function create_guis(player) 
	if player.gui.top.curscore then
		player.gui.top.curscore.destroy()
	end
	if player.gui.top.leaderboard_scores then
		player.gui.top.leaderboard_scores.destroy()
	end
	if player.gui.top.admin then
		player.gui.top.admin.destroy()
	end
	player.gui.top.add{name = "curscore", type = "frame", direction = "vertical", caption="This Round"}
	player.gui.top.curscore.add{type="flow", name="panel", direction="vertical"}
	local g = player.gui.top.add{name = "leaderboard_scores", type = "frame", direction = "vertical", caption="Leaderboard"}
	g.add{type="flow", name="panel", direction="vertical"}
	for i=1,global.display_count do
		g.panel.add{name = i, type = "label"}
	end
	if player.admin then
		g = player.gui.top.add{type="frame", name="admin"}
		g.add{type="flow", name="alwayson", direction="vertical"}
		g.alwayson.add{type="button", name="showbutton", caption="Config  "}
		g.alwayson.add{type="button", name="restartbutton", caption="Restart"}
		g = g.add{type="flow", name = "adminpanel", direction = "vertical"}
		gp = g.add{type="flow", name="admintable"}
		local gl = gp.add{type="flow", name="cfglabel", direction="vertical"}
		local gt = gp.add{type="flow", name="cfgtext", direction="vertical"}
		-- max kill
		gl.add{type="label",name="kill_limit",caption="Kill limit per round"}
		gt.add{type="textfield",name="kill_limit",text=global.kill_limit}
		-- players on leaderboard_scores
		gl.add{type="label",name="count",caption="Leaderboard display count"}
		gt.add{type="textfield",name="count",text=global.display_count}
		apply_style({gl.kill_limit, gl.count}, cfglabel_style)
		apply_style({gt.kill_limit, gt.count}, cfgtext_style)
		for tier,_ in pairs(global.chest_loot_table) do 	
			-- chest init delay
			gl.add{type="label",name=tier .. "_initdelay", caption=tier .. " initial spawn delay"}
			gt.add{type="textfield", name=tier .. "_initdelay", text=global.chest_initial_delay[tier]}
			-- chest respawn delay
			gl.add{type="label", name=tier .. "_respawndelay", caption=tier .. " respawn delay"}
			gt.	add{type="textfield", name=tier .. "_respawndelay", text=global.chest_respawn_delay[tier]}		
			apply_style({gt[tier .. "_respawndelay"], gt[tier .. "_initdelay"]}, cfgtext_style)
			apply_style({gt[tier .. "_respawndelay"], gt[tier .. "_initdelay"]}, cfglabel_style)
		end
		--update button
		g.add{type="button", name="updatebutton", caption="Update cfg"}
		apply_style(g.updatebutton, updatebutton_style)
		g.style.visible = false
	end
	update_guis(true)
end

-- update gui process
function update_guis(full)	
	for _,player in pairs(game.connected_players) do 
		--update leaderboard_scores
		if full then
			local sco = sort_scores(global.leaderboard_scores, math.min(global.display_count, global.unique_player_count))
			for k,v in pairs(sco) do
				s = k .. ". " .. v.name .. " : " .. v.score
				player.gui.top.leaderboard_scores.panel[k].caption = s
			end
		end
		--del all curscore labels
		for _, g in pairs(player.gui.top.curscore.panel.children_names) do
			player.gui.top.curscore.panel[g].destroy()
		end		
		--add labels
		for k,v in pairs(global.current_round_scores) do 		
			local s = k .. " : " .. v
			player.gui.top.curscore.panel.add{name=k, type = "label", caption=s}
		end
	end
end

-- popup between rounds
function end_round_popup(player)
	if player.gui.center.endframe then
		player.gui.center.endframe.destroy()
	end
	local g = player.gui.center.add{name="endframe", type="frame", direction = "vertical", caption="Round finished"}
	g.add{name="panelwin", type="flow", direction="vertical"}
	local l = g.panelwin.add{type="label", caption=player.name==global.winner and "You Win" or "You loose"}
	apply_style(l,win_label_style)
	g.add{name="panel", type="flow", direction="vertical"}
	local sco = sort_scores(global.current_round_scores, #game.connected_players)
	for k,v in pairs(sco) do
		local s = k .. " - " .. v.name .. " : " .. v.score
		g.panel.add{type = "label", caption=s}
	end
end

-----------------------
-- Player management --
-----------------------

-- equip & teleport player
function prepare(player)
	for i=1, #player_default_items do
		player.insert(player_default_items[i])
	end
	teleport(player)
end

function teleport(player)
	math.randomseed(game.tick)
	if random_spawns then
		-- generate new spawn coordinates up to random_spawn_limit
		local pos = #global.generated_spawn_list == global.random_spawn_limit and global.generated_spawn_list[math.random(global.random_spawn_limit)] or generate_new_spawn()
		player.teleport(pos, player.surface)
	else	
		-- pick random coordinates from the static spawn table
		player.teleport(global.spawn_list[math.random(#global.spawn_list)],player.surface)
	end
end

--------------------
-- Map & loots --
--------------------

-- test random coordinates, if valid spawn found, add them to the generated spawn table and returns them
function generate_new_spawn()
	local randx, randy
		repeat
			randx=math.random(arena_box[1][1],arena_box[2][1])
			randy=math.random(arena_box[1][2],arena_box[2][2])
		until game.surfaces[1].can_place_entity{name="player", position={randx,randy}} and not isin(game.surfaces[1].get_tile(randx,randy).name, random_spawn_exclude_tiles)	
		table.insert(global.generated_spawn_list,{randx,randy})
		return {randx,randy}
end

-- will add any empty chest to the refill queue if not already
function chest_respawn_tick_register(tick) 
	for tier, chests in pairs(global.chest_table) do
		for _,chest in pairs(chests) do
			-- detect which chest has been emptied and not already scheduled for respawn
			if not chest.has_items_inside() and global.activated_chests[chest.name] and not isin(chest, global.chest_respawn_ticks) then
				global.chest_respawn_ticks[tick + 6 + 60 * global.chest_respawn_delay[tier]] = chest
			end
		end		
	end
end

-- set the next tick where all chest will be filled (after a new_round)
function chest_initial_tick_register(tick) 
	for tier,_ in pairs(global.chest_loot_table) do
		global.chest_initial_tick[global.chest_initial_delay[tier]*60 + tick + 6] = tier
	end
end

-- adjust the loot table to replace probability by duplicates of the entry
--[[ exemple : 
	chest_loot_table = {tier1 = {{item = {name="land-mine",count=1}, probability = 3}, {item = {name="shotgun-shell",count=2},probability = 1}} becomes	
	chest_loot_table = {tier1 = {{name="land-mine",count=1}, {name="land-mine",count=1}, {name="land-mine",count=1}, {name="shotgun-shell",count=2}}
--]]
function format_loot_table()
	local newitemslist = {}
	for tier, itemlist in pairs(global.chest_loot_table) do
		newitemslist[tier] = {}
		for _, itemset in pairs(itemlist) do
			for i=1, itemset.probability do
				table.insert(newitemslist[tier], itemset.item)
			end		
		end		
	end
	global.chest_loot_table = newitemslist
end

-- remove player's land mines/turrets or everyone's if no player specified
function clean_map(player)
	local a
	if player then
		a = game.surfaces[1].find_entities_filtered{name="land-mine", force=player.force, area=arena_box}
		b = game.surfaces[1].find_entities_filtered{name="gun-turret", force=player.force, area=arena_box}
	else
		a = game.surfaces[1].find_entities_filtered{name="land-mine", area=arena_box}
		b = game.surfaces[1].find_entities_filtered{name="gun-turret", area=arena_box}
	end
	for _,e in ipairs(a) do
		e.destroy()
	end
	for _,e in ipairs(b) do
		e.destroy()
	end
end

-----------
-- Utils --
-----------

function apply_style (guis, style)
	if guis.type then -- test if it's a gui
		for k,v in pairs(style) do
			guis.style[k]=v
		end		
	else -- or an array of guis
		for _,g in ipairs(guis) do
			for k,v in pairs(style) do
				g.style[k]=v
			end
		end	
	end
end

function sort_scores(tab,size)
	local result = {}
	local temp = {}
	local count = 1
	for k,v in pairs(tab) do
		table.insert(temp,{name=k, score=v})
	end
	while(count <= size) do
		local maxscore = -1
		local maxname
		local maxindex
		for k,v in pairs(temp) do
			if v.score>maxscore then 
				maxscore = v.score
				maxname = v.name
				maxindex = k
			end		
		end
		count=count+1
		table.insert(result,{name=maxname, score=maxscore})
		table.remove(temp,maxindex)		
	end
	return result
end

function isin(item,list)
	for _, i in pairs(list) do 
		if item == i then return true end	
	end
	return false
end

------------
-- Styles --
------------

win_label_style = {
	font = "default-large-bold",
	font_color = {r=1}
}

cfgtext_style = {
	maximal_width = 40,
	minimal_height = 20,
	maximal_height = 20
}

cfglabel_style = {
	minimal_height = 20
}

updatebutton_style = {
	minimal_width = 150
}